#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

// This signal handler runs every time a child process terminates.
// In fact a SIGCHLD signal is generated and sent to the parent process
// which catches the signal and runs this handler() code.
void handler(int sig)
{
  pid_t t;
  int status;
  
  // Wait until the child exits
  t=wait(&status);
  printf("Child %d exited with status %d\n",t,WEXITSTATUS(status));
}

int main(int argc, char *argv[])
{
  int processes;
  int i,alive;
  pid_t *pids;
  
  if (argc!=2) {
    fprintf(stderr,"\nUsage\t%s <nchilds>\n",argv[0]);
    exit(1);
  }
  
  // The first arg on command line is the number of child to be forked
  processes=atoi(argv[1]);
  
  // Register the signal handler for the SIGCHLD signal
  signal(SIGCHLD,handler);
  
  // Allocate an array of <processes> items of type pid_t 
  if (!(pids=malloc(processes*sizeof(pid_t)))) {
    perror("malloc");
    exit(1);
  }
  
  // Loop to fork the child processes
  for (i=0; i<processes; i++) {
    int t,s;
    
    switch (pids[i]=fork())  {
    case 0: // Child process
      s=getpid();

      // Set the seed for the random number returned by rand(): 
      //  the seed is the pid of the process,
      //  so every child process gets a different random number
      srand(s);
      t=(double)rand()/RAND_MAX*10+1; // Race!!!
      printf("Child %d: sleeping for %d sec.\n",s,t);
      sleep(t);

      // Malloced memory is inherited, so must be freed!
      free(pids);
      exit(0);
    case -1: // Error from fork
      perror("fork");
      exit(1);
    default: // Parent process: 
      // execute all the iterations of this for-loop and then 
      // code continues with the following do-loop
      break;
    }
  }
  
  // Loop to check periodically the child status
  do {
    alive=0;
    printf("sleeping\n");
    // Sleep for a while and then check in a sequence if every child process
    //  is still alive or is dead. 
    // The do-loop exits when all the child processes are dead.
    sleep(1);
    for (i=0; i<processes; i++) {
      if (kill(pids[i],0)) {
	// When kill(pid,0) returns -1, that is error, it means
	//  that the pid does not exist (to be precise we should also
	//  check that errno is ESRCH, see man 2 kill)
	printf("%d is dead\n",pids[i]);
      }
      else { 
	// When kill(pid,0) returns 0, that is no error, it means
	//  that the pid exists, so we assume that the associated process 
	//  is alive (...but it might be also a zombie!)
	printf("%d is alive\n",pids[i]);
	alive=1;
      }
    }
  } while (alive);
  
  // Free allocated memory
  free(pids);

  exit(0);
}
